/*
 *  Copyright (C) 2021 Texas Instruments Incorporated
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* This example demonstrates the UART RX and TX operation by echoing char
 * that it recieves in blocking, interrupt mode of operation.
 * When user types 'quit', the application ends.
 */

#include <string.h>
#include <kernel/dpl/DebugP.h>
#include "ti_drivers_config.h"
#include "ti_drivers_open_close.h"
#include "ti_board_open_close.h"

#include "model.h"

#include <time.h>


#define SAMPLES_PER_STRING 500 // 3   // NEW
#define PREDICTIONS_PER_OUTPUT 50 // 3   // NEW

//#define APP_UART_Rx_BUFSIZE      (95U) // 15x6 + 5 = 95 //(76U)  // (33U) // (11U)  // ADM
#define APP_UART_Rx_BUFSIZE      (47999U) //(287U) // 15x6x3 + (6-1)x3 + (3-1) = 287 //(76U)  // (33U) // (11U)  // ADM // NEW
//#define APP_UART_Rx_BUFSIZE      (185U) // 30x6 + 5 = 185 //(76U)  // (33U) // (11U)  // ADM
//#define APP_UART_Rx_BUFSIZE      (905U) // 150x6 + 5 = 185 //(76U)  // (33U) // (11U)  // ADM
#define APP_UART_Tx_BUFSIZE      (400U) //(24U) //8*3= 24 // (8U)  // (6U)   // ADM

#define ALPHA 0.00004

uint8_t gUartReceiveBuffer[APP_UART_Rx_BUFSIZE];    // ADM
uint8_t gUartTransmitBuffer[APP_UART_Tx_BUFSIZE];    // ADM

volatile uint32_t gNumBytesRead = 0U, gNumBytesWritten = 0U;

#define APP_UART_ASSERT_ON_FAILURE(transferOK, transaction) \
    do { \
        if((SystemP_SUCCESS != (transferOK)) || (UART_TRANSFER_STATUS_SUCCESS != transaction.status)) \
        { \
            DebugP_assert(FALSE); /* UART TX/RX failed!! */ \
        } \
    } while(0) \


float low_pass_filter(float lpf_alpha, float data_out_buf, int reset)
    {
        static float previous_data_out_buf = 0;
        if (reset == 1){
            previous_data_out_buf = 0;
        }
        float lowpass_data = lpf_alpha*data_out_buf + (1-lpf_alpha)*previous_data_out_buf;
        previous_data_out_buf = lowpass_data;

        return lowpass_data;
    }



void uart_echo(void *args)
{
        int32_t          transferOK;
        UART_Transaction trans_Rx, trans_Tx;

        Drivers_open();
        Board_driversOpen();

        DebugP_log("[UART] Echo example started ...\r\n");

        UART_Transaction_init(&trans_Rx);   // ADM
        UART_Transaction_init(&trans_Tx);   // ADM

        /*--------- Read APP_UART_Rx_BUFSIZE chars - data in string ---------*/
        //gNumBytesRead = 0U;
        //trans_Rx.buf   = &gUartReceiveBuffer[0U];
        //trans_Rx.count = APP_UART_Rx_BUFSIZE;

        /*--------- Send APP_UART_Tx_BUFSIZE chars - data out string ---------*/
        //gNumBytesWritten = 0U;
        //trans_Tx.buf   = &gUartTransmitBuffer[0U];
        //trans_Tx.count = APP_UART_Tx_BUFSIZE;


        /*
        int NUMBER_OF_OUTPUT_DIGITS = 7 + 1;
        char output_buf_new[NUMBER_OF_OUTPUT_DIGITS];
        output_buf_new[0] = '1';
        output_buf_new[1] = '2';
        output_buf_new[2] = '3';
        output_buf_new[3] = '4';
        output_buf_new[4] = '5';
        output_buf_new[5] = '\r';
        output_buf_new[6] = '\n';

        for (int i = 0; i < NUMBER_OF_OUTPUT_DIGITS; i++) {
            strncpy(trans_Tx.buf+i, &output_buf_new[i], 1U);
        }
        while(1)
        {
            transferOK = UART_write(gUartHandle[CONFIG_UART_CONSOLE], &trans_Tx);
            APP_UART_ASSERT_ON_FAILURE(transferOK, trans_Tx);
        }
        */



        //int counter = 0;

        int NUMBER_OF_DIGITS = 15 + 1; // 1 digit accounts for the decimal point and 1 digit for end string '\0'
        int NUMBER_OF_OUTPUT_DIGITS = 8U + 0; // 1 digit accounts for the decimal point and 1 digit for end string '\0'

        float data_in_buf[IMAI_DATA_IN_COUNT*SAMPLES_PER_STRING] = {0}; // NEW
        float data_out_buf[IMAI_DATA_OUT_COUNT] = {0}; // NEW

        char temp_buf[NUMBER_OF_DIGITS];
        //////////////////////////temp_buf[NUMBER_OF_DIGITS - 1] = '\0';

        int output_index = 0;
        int reset = 0;

        char output_buf[NUMBER_OF_OUTPUT_DIGITS*PREDICTIONS_PER_OUTPUT];  // NEW
        //output_buf[0] = '0';
        //output_buf[1] = '0';
        //for(int i=2; i < NUMBER_OF_OUTPUT_DIGITS - 1; i++){
        //    output_buf[i] = '0';
        //}
        //output_buf[10] = '\0';

        while (1){
            /*--------- Read APP_UART_Rx_BUFSIZE chars - data in string ---------*/
           gNumBytesRead = 0U;
           trans_Rx.buf   = &gUartReceiveBuffer[0U];
           trans_Rx.count = APP_UART_Rx_BUFSIZE;
            /*--------- Read APP_UART_Rx_BUFSIZE chars - data in string ---------*/
            transferOK = UART_read(gUartHandle[CONFIG_UART_CONSOLE], &trans_Rx);
            APP_UART_ASSERT_ON_FAILURE(transferOK, trans_Rx);


            /*--------- Read CSV string ---------*/
            //DebugP_log("-------------------------------------------------\n");
            //DebugP_log("Reading CSV String and Converting to Float Array \n");
            //DebugP_log("-------------------------------------------------\n");

            //DebugP_log("\n Data %s \n", gUartReceiveBuffer);

            if (gUartReceiveBuffer[0] == 'r')
            {
                IMAI_init();
                reset = 1;
                output_index = 0;
                //DebugP_log("Reset... \n");
                continue;
            }


            /*--------- Convert string to float array ---------*/
            char *temp_ptr;
            int i = 0;
            int j = 0;
            int k = 0;
            while (i < APP_UART_Rx_BUFSIZE) {
                //DebugP_log("The character in position %d = %c \n", i, gUartReceiveBuffer[i]);
                while (gUartReceiveBuffer[i] != ',' && k < (NUMBER_OF_DIGITS - 1)) {
                    temp_buf[k] = gUartReceiveBuffer[i];
                    i++;
                    k++;
                }
                while (k < (NUMBER_OF_DIGITS - 1)){
                    temp_buf[k] = '0';
                    k++;
                }
                if (j < IMAI_DATA_IN_COUNT*SAMPLES_PER_STRING){  // NEW
                    data_in_buf[j] = strtod(temp_buf, &temp_ptr);
                    j++;
                    k = 0;
                }
                i++;
            }

            /*
            for(int i=0; i<IMAI_DATA_IN_COUNT*SAMPLES_PER_STRING; i++)   // NEW
            {
                //DebugP_log("Data In %d %c \n", i, gUartReceiveBuffer[i]);
                DebugP_log("[%d] %f ", i, data_in_buf[i]);   // NEW
            }
            DebugP_log("\n");
            */

            float data_in_buf_enqueue[IMAI_DATA_IN_COUNT] = {0}; // NEW
            float data_out_buf_dequeue[IMAI_DATA_OUT_COUNT*SAMPLES_PER_STRING] = {0};  // NEW

            int z = 0;

            while(z < IMAI_DATA_IN_COUNT*SAMPLES_PER_STRING){                // NEW
                for(int j = 0; j < IMAI_DATA_IN_COUNT; j++){
                    data_in_buf_enqueue[j] = data_in_buf[z+j];
                }
                /*
                for(int k=0; k<IMAI_DATA_IN_COUNT; k++)   // NEW
                {
                    DebugP_log("[%d] %f ", k, data_in_buf_enqueue[k]);   // NEW
                }
                DebugP_log("\n");
                */

                //time_t t0 = time(0);
                int status = IMAI_enqueue(data_in_buf_enqueue);    //passing data to the model  // NEW
                //time_t t1 = time(0);
                //double datetime_diff_ms = difftime(t1, t0) * 1000.;
                //DebugP_log("Enqueue %f \n", datetime_diff_ms);

                //DebugP_log("Status %d \n", status);

                //time_t t2 = time(0);
                //int dequeue_status = 0 //IMAI_dequeue(data_out_buf);
                //time_t t3 = time(0);
                //double datetime_diff_ms_dequeue = difftime(t3, t2) * 1000.;
                //DebugP_log("Dequeue %f \n", datetime_diff_ms_dequeue);

                if(IMAI_dequeue(data_out_buf) == 0)
                {
                    //DebugP_log("Res %f \n", data_out_buf[0]);
                    //DebugP_log("Res %f \n", gUartTransmitBuffer[0]);

                    //counter++;
                    //if(counter % 10 == 0)
                    //    DebugP_log("Count %d \n", counter);

                    /*--------- Convert float array to string ---------*/
                    //char output_buf[NUMBER_OF_OUTPUT_DIGITS*SAMPLES_PER_STRING];
                    //output_buf[0] = '0';
                    //output_buf[1] = '0';
                    //for(int i=2; i < NUMBER_OF_OUTPUT_DIGITS - 1; i++){
                    //    output_buf[i] = '0';
                    //}
                    //output_buf[10] = '\0';
                    float out_temp = low_pass_filter(ALPHA, data_out_buf[0], reset);
                    data_out_buf[0] = out_temp;
                    reset = 0;

                    if (output_index < NUMBER_OF_OUTPUT_DIGITS*PREDICTIONS_PER_OUTPUT){
                        sprintf(output_buf+output_index,"%d.%06u", (int) data_out_buf[0], (int) (fabs(((data_out_buf[0] - (int) data_out_buf[0] ) * 1000000))));
                        output_index = output_index + NUMBER_OF_OUTPUT_DIGITS;
                    }
                    else {
                        //DebugP_log("F to S = %s \n", output_buf);
                        // OR
                        //for(int i=0;i<NUMBER_OF_OUTPUT_DIGITS;i++){
                        //    DebugP_log("Char = %c \n", output_buf[i]);
                        //}

                        output_index = 0;

                        /*--------- Send APP_UART_Tx_BUFSIZE chars - data out string ---------*/
                        gNumBytesWritten = 0U;
                        trans_Tx.buf   = &gUartTransmitBuffer[0U];
                        trans_Tx.count = APP_UART_Tx_BUFSIZE;

                        for (int i = 0; i < NUMBER_OF_OUTPUT_DIGITS*PREDICTIONS_PER_OUTPUT; i++) {
                            strncpy(trans_Tx.buf+i, &output_buf[i], 1U);
                            }

                        transferOK = UART_write(gUartHandle[CONFIG_UART_CONSOLE], &trans_Tx);
                        APP_UART_ASSERT_ON_FAILURE(transferOK, trans_Tx);

                        //for(int i=0; i<APP_UART_Tx_BUFSIZE; i++)
                        //    gUartTransmitBuffer[i] = '0';
                    }
                }
                //for(int i=0; i<APP_UART_Rx_BUFSIZE; i++)
                //    gUartReceiveBuffer[i] = '0';

                z = z + IMAI_DATA_IN_COUNT;  // NEW
            }
        }


        DebugP_log("All tests have passed!!\r\n");

        Board_driversClose();
        Drivers_close();

        return;
}
